/*-
 * Copyright (C)2017..2018 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2017..2018 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: func.c,v 1.28 2018/08/13 21:25:10 babolo Exp $"

#define BLIN_COMPAT      3
#define Bpars_COMPAT     3
#define MIFE_COMPAT      4
#define RECOBE_COMPAT    4
#define PGOBLIN_COMPAT   2
#define PGOBLIN_INTERNAL 1

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sysexits.h>
#include <strings.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include <multilar.h>
#include <mife.h>
#include "pgoblin.h"

int
/************************************************************************
 **                                                                    **/
pgoblin_xwrt(pgoblin_main *options, void *buf, int len) {             /**
 **                                                                    **
 ************************************************************************/
    int wrt;

    errno = 0;
    if  (len == (wrt = mife_writ(options->xtrace, buf, len))) {
        wrt = 0;
    }
    return(wrt);
}

static const u_char *
/************************************************************************
 **                                                                    **/
winpoint(mife_descriptor *md, off_t offset, ssize_t trd) {            /**
 **                                                                    **
 ************************************************************************/
    const u_char    *p = NULL;
    ssize_t          rdr;

    errno = 0;
    if  (trd > (rdr = mife_read(md, trd, offset))) {
        if  (!!mife_ctlsize(md)) {
            ifBLIN_QW0("mife_read %"BLIN_D" < %"BLIN_D, rdr, trd);
            if  (rdr >= 0) errno = EFTYPE;
        }
    } else if (!(p = mife_pointer(md))) {
        ifBLIN_QW0("mife_pointer");
    }
    return(p);
}

int
/************************************************************************
 ************************************************************************
 **                                                                    **/
pgoblin_breg(pgoblin_command cmd, pgoblin_nr rn) {                    /**
 **                                                                    **
 ************************************************************************
 **      cmd   rn.     **
 **    cmd  rn  ,   < 0 **
 ************************************************************************
 ************************************************************************/
    int j;

    j = -1;
    if  (pgoblin.a[cmd].flags & pgoblin.p[rn].sym) {
        /*  a[cmd].flags    [rn] */
        j = (pgoblin.p[rn].bin < 0)
            /*      */
          ? (pgoblin.a[cmd].flags >> ~pgoblin.p[rn].bin) & pgoblin.p[rn].mask
            /*      */
          : pgoblin.p[rn].bin
        ;
        if  (!(pgoblin.a[cmd].ch[j] & pgoblin.p[rn].val) && !!pgoblin.a[cmd].ch[j]) {
            /*                            *
             *     ,  -     . *
             *    0 j ,   j  ,      *
             *        j                    *
             */
            j = -2;
    }   }
    return(j);
}

static void
/**********************************************************************
 **                                                                  **/
putnum(pgoblin_main *options, u_int32_t num) {                      /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    u_char  buff[10];
    ssize_t wrt;
    int     l;

    l = recobe_cstore(buff, num);
    if  (!!(wrt = pgoblin_xwrt(options, buff, l))) {
        ifBLIN_QW0("mife_writ %d <> %d", wrt, l);
}   }
#   undef blin_internal_flags

static void
/**********************************************************************
 **                                                                  **/
putext(pgoblin_main *options, char *text, size_t length, int bin) { /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    char    null = 0xFF;
    ssize_t strl;
    ssize_t wrt;

    if  (!text) {
        if  (!!(wrt = pgoblin_xwrt(options, &null, 1))) {
            ifBLIN_QW0("mife_writ %d <> 1", wrt);
        }
    } else {
        if (!!bin || !!length) {
           strl = length;
        } else {
           strl = strlen(text);
        }
        if  (strl > 0x000000007FFFFFFFLL) {
            ifBLIN_QX0("string too big %"BLIN_X, strl);
        } else {
            putnum(options, strl);
            if  (!!(wrt = pgoblin_xwrt(options, text, strl))) {
                ifBLIN_QW0("mife_writ %"BLIN_D" <> %"BLIN_D, wrt, strl);
}   }   }   }
#   undef blin_internal_flags

int
/**********************************************************************
 **********************************************************************
 **                                                                  **/
pgoblin_wident(pgoblin_main *options, int num, char *entry) {       /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    int            ex = EX_OK;
    u_char         buf[PGOBLIN_GOSHORT];
    size_t         len;

    if  (0 <= options->xtrace) {
        buf[0] = 0xA0;
        buf[1] = num;
        recobe_store(&buf[PGOBLIN_HEDLN], PGOBLIN_PIDLN, getpid());
        len = strnlen(entry, PGOBLIN_STRING_ID_LEN);
        if  (0 > (options->xtrace = open(options->ftrace, O_WRONLY | O_APPEND | O_EXLOCK))) {
            ifBLIN_QW0("open(%s, WRONLY|APPEND|EXLOCK)", options->ftrace);
        } else {
            int wrt;

            if  (!!(wrt = pgoblin_xwrt(options, buf, PGOBLIN_GOSHORT))) {
                ifBLIN_QW0("pgoblin_xwrt %d <> %d", wrt, PGOBLIN_GOSHORT);
            }
            putext(options, entry, len, 1);
            entry += PGOBLIN_STRING_ID_LEN;
            len = strnlen(entry, PGOBLIN_STRING_ID_LEN);
            if  (len < PGOBLIN_STRING_ID_LEN) {
                putext( options
                      , entry + len + 1
                      , strnlen(entry + len + 1, PGOBLIN_STRING_ID_LEN - len - 1)
                      , 1
                      );
            } else {
                putext(options, NULL, 0, 0);
            }
            putext(options, entry, len, 1);
            close(options->xtrace);
        }
        ifBLIN_QX3( "- %d", ex);
    }
    return(ex);
#   undef blin_internal_flags
}

void
/**********************************************************************************
 **********************************************************************************
 **                                                                              **/
pgoblin_toexe(int cmd, pgoblin_exenv *exenv, pgoblin_nr rn[pgoblin.symparam]) { /**
 **                                                                              **
 **********************************************************************************
 **********************************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    if  (!!exenv->options->ftrace) {
        u_char         buf[PGOBLIN_GOEXECU];
        int            binreg;
        int            symreg;
        pgoblin_exenv *c;
        int32_t        d;
        int32_t        s;

        for (s = 0, c = exenv; c; c = c->stat) s++;
        for (d = 0, c = exenv; c; c = c->dyna) d++;
        bzero(buf, PGOBLIN_GOEXECU);
        buf[0] = (cmd >> 8) | 0x0C0;
        buf[1] = cmd & 0x0FF;
        recobe_store(&buf[PGOBLIN_HEDLN], PGOBLIN_PIDLN, getpid());
        for (symreg = 0; symreg < pgoblin.symparam; symreg++) {
            if  (0 <= (binreg = pgoblin_breg(cmd, symreg))) {
                buf[PGOBLIN_GOSHORT + binreg] = rn[symreg];
        }   }
        recobe_store(&buf[PGOBLIN_GOCMD], sizeof(int32_t), s);
        recobe_store(&buf[sizeof(int32_t) + PGOBLIN_GOCMD], sizeof(int32_t), d);
        recobe_store(&buf[sizeof(int32_t) * 2 + PGOBLIN_GOCMD], sizeof(int32_t), exenv->ppoint);
        if  (0 > (exenv->options->xtrace = open( exenv->options->ftrace
                                               , O_WRONLY | O_APPEND | O_EXLOCK
            )     )                            ) {
            ifBLIN_QW0("open(%s, WRONLY|APPEND|EXLOCK)", exenv->options->ftrace);
        } else {
            int wrt;

            if  (!!(wrt = pgoblin_xwrt(exenv->options, buf, PGOBLIN_GOEXECU))) {
                ifBLIN_QW0("pgoblin_xwrt %d <> %d", wrt, PGOBLIN_GOEXECU);
            }
            close(exenv->options->xtrace);
}   }   }
#   undef blin_internal_flags

int
/**********************************************************************
 **********************************************************************
 **                                                                  **/
pgoblin_cmdexe(int cmd, pgoblin_exenv *exenv, pgoblin_nr *rn) {     /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    u_char         buf[PGOBLIN_GOEXECU];
    char           null = 0xFF;
    u_int8_t       binreg;
    u_int8_t       symreg;
    ssize_t        wrt;
    int            ex;

    ifBLIN_QX3( "+%s %c%c%c%c%c%c %08X"
              , pgoblin.a[cmd].str
              , pgoblin_regn[rn[0]]
              , pgoblin_regn[rn[1]]
              , pgoblin_regn[rn[2]]
              , pgoblin_regn[rn[3]]
              , pgoblin_regn[rn[4]]
              , pgoblin_regn[rn[5]]
              , pgoblin.a[cmd].flags
              );
    if  (!!exenv->options->ftrace) {
        {   pgoblin_exenv *c;
            int32_t        d;
            int32_t        s;

            for (s = 0, c = exenv; c; c = c->stat) s++;
            for (d = 0, c = exenv; c; c = c->dyna) d++;
            bzero(PGOBLIN_GOSHORT + buf, pgoblin.binparam);
            buf[0] = (cmd >> 8) | 0x0E0;
            buf[1] = cmd & 0x0FF;
            recobe_store(&buf[PGOBLIN_HEDLN], PGOBLIN_PIDLN, getpid());
            for (symreg = 0; symreg < pgoblin.symparam; symreg++) {
                if  (0 <= (binreg = pgoblin_breg(cmd, symreg))) {
                    buf[PGOBLIN_GOSHORT + binreg] = rn[symreg];
            }   }
            recobe_store(&buf[PGOBLIN_GOCMD], sizeof(int32_t), s);
            recobe_store(&buf[sizeof(int32_t) + PGOBLIN_GOCMD], sizeof(int32_t), d);
            recobe_store(&buf[sizeof(int32_t) * 2 + PGOBLIN_GOCMD], sizeof(int32_t), exenv->ppoint);
        }
        if  (0 > (exenv->options->xtrace = open( exenv->options->ftrace
                                               , O_WRONLY | O_APPEND | O_EXLOCK
            )     )                            ) {
            ifBLIN_QW0("open(%s, WRONLY|APPEND|EXLOCK)", exenv->options->ftrace);
        } else {
            if  (!!(wrt = pgoblin_xwrt(exenv->options, buf, PGOBLIN_GOEXECU))) {
                ifBLIN_QW0("mife_writ %d <> %d", wrt, PGOBLIN_GOEXECU);
            }
            for (binreg = 0; binreg < pgoblin.binparam; binreg++) {
                pgoblin_io *io = &exenv->options->io[buf[PGOBLIN_GOSHORT + binreg] & 0x03F];

#ifndef NOTRACE_PATHPARM
                if  (PGOBLIN_PATHPARM & pgoblin.a[cmd].ch[binreg]) {
                ;   ifBLIN_QX4("[%d]PATHPARM", binreg);
                ;   putext(exenv->options, io->path, 0, 0);
                }
#endif
#ifndef NOTRACE_TEXTPARM
                if  (PGOBLIN_TEXTPARM & pgoblin.a[cmd].ch[binreg]) {
                ;   ifBLIN_QX4("[%d]TEXTPARM", binreg);
                ;   putext(exenv->options, io->text, io->length, io->flags & PGOBLIN_BINPARM);
                }
#endif
#ifndef NOTRACE_MIFEDES
                if  (PGOBLIN_MIFEDES & pgoblin.a[cmd].ch[binreg]) {
                ;   ifBLIN_QX4("[%d]MIFEDES", binreg);
                }
#endif
#ifndef NO_PQRESULT
                if  (PGOBLIN_PQRESULT & pgoblin.a[cmd].ch[binreg]) {
                    int     row;
                    int     col;
                    int32_t v;

                ;   ifBLIN_QX4("[%d]PQRESULT", binreg);
                ;   if  (!io->pq) {
                ;       if  (!!(wrt = pgoblin_xwrt(exenv->options, &null, 1))) {
                ;           ifBLIN_QW0("mife_writ %d <> 1", wrt);
                ;       }
                ;   } else {
                ;       if  (0 > (v = pgoblin_db_resinfo(PGOBLIN_Ntuples, io->pq))) {
                ;           ifBLIN_QW0("resinfo Ntuples");
                ;       } else {
                ;           putnum(exenv->options, v);
                ;       }
                ;       if  (0 > (v = pgoblin_db_resinfo(PGOBLIN_Nfields, io->pq))) {
                ;           ifBLIN_QW0("resinfo Nfields");
                ;       } else {
                ;           putnum(exenv->options, v);
                ;       }
                ;       for (row = 0; row < pgoblin_db_resinfo(PGOBLIN_Ntuples, io->pq); ++row) {
                ;           for (col = 0; col < v; ++col) {
                ;               if  (pgoblin_db_valinfo(PGOBLIN_IsNull, io->pq, row, col)) {
                ;                   putext(exenv->options, NULL, 0, 0);
                ;               } else {
                ;                   putext( exenv->options
                                          , pgoblin_db_getvalue(io->pq, row, col)
                                          , pgoblin_db_valinfo(PGOBLIN_Length, io->pq, row, col)
                                          , 0
                                          );
                }   }   }   }   }
#endif
#ifndef NOTRACE_CONNECT
                if  (PGOBLIN_CONNECT & pgoblin.a[cmd].ch[binreg]) {
                ;   ifBLIN_QX4("[%d]CONNECT", binreg);
                }
#endif
#ifndef NOTRACE_JOB
                if  (PGOBLIN_JOB & pgoblin.a[cmd].ch[binreg]) {
                ;   ifBLIN_QX4("[%d]JOB", binreg);
                }
#endif
                if  (PGOBLIN_STYLE & pgoblin.a[cmd].ch[binreg]) {
#ifndef NOTRACE_STYLE
                ;   ifBLIN_QX4("[%d]STYLE", binreg);
#endif
        }   }   }
        close(exenv->options->xtrace);
    }
    ex = (*(pgoblin.a[cmd].car))(exenv, rn);
    ifBLIN_QX3( "- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

void *
/**********************************************************************
 **********************************************************************
 **                                                                  **/
pgoblin_getr(pgoblin_main *options, pgoblin_nr rn, int reg) {       /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    switch(reg) {
    case PGO_CSTY: return(&(options->style[rn]));
    case PGO_CJOB: return(&(options->job[rn]));
    case PGO_CCON: return(&(options->conn[rn]));
    default:       return(&(options->io[rn]));
#   undef blin_internal_flags
}   }

void
/*****************************************************************************
 *****************************************************************************
 **                                                                         **/
pgoblin_go( void *x           /*  (f > 0)   (f < 0)     */
/**/      , int f             /*    pgoblin_syntax.ch[]  *-1 */
/**/      , const char *file  /*                            */
/**/      , const char *func  /*                                   */
/**/      , int line          /*                        */
/**/      ) {                                                              /**
 **                                                                         **
 *****************************************************************************
 *****************************************************************************/
#   define blin_internal_flags (pgoblin_options->flags & BLIN_MASK)
    if  (!!pgoblin_options->ftrace) {
        u_char    buf[PGOBLIN_GOSHORT];
        u_int64_t d;
        u_int64_t p;
        u_int64_t q;
        int       i;

        if  (0 < f) {
            switch(f) {
            case PGOBLIN_STYLE   :
            case PGOBLIN_JOB     :
            case PGOBLIN_CONNECT :
            case PGOBLIN_OUTSET  :
            case PGOBLIN_PQRESULT:
            case PGOBLIN_MIFEDES :
            case PGOBLIN_TEXTPARM:
            case PGOBLIN_PATHPARM:
                buf[0] = f;
                buf[1] = (u_int64_t)x;
                break;
            default              :
                buf[0] = 0;
                buf[1] = f;
            }
        } else {
            switch(-f) {
            case PGOBLIN_STYLE   :
                p = (u_int64_t)&pgoblin_options->style[0];
                q = (u_int64_t)&pgoblin_options->style[1];
                break;
            case PGOBLIN_JOB     :
                p = (u_int64_t)&pgoblin_options->job[0];
                q = (u_int64_t)&pgoblin_options->job[1];
                break;
            case PGOBLIN_CONNECT :
                p = (u_int64_t)&pgoblin_options->conn[0];
                q = (u_int64_t)&pgoblin_options->conn[1];
                break;
            case PGOBLIN_OUTSET  :
            case PGOBLIN_PQRESULT:
            case PGOBLIN_MIFEDES :
            case PGOBLIN_TEXTPARM:
            case PGOBLIN_PATHPARM:
                p = (u_int64_t)&pgoblin_options->io[0];
                q = (u_int64_t)&pgoblin_options->io[1];
                break;
            default              :
                buf[0] = 0;
                buf[1] = -f;
                goto out;
            }
            d = ((u_int64_t)x - p) / (q - p);
            if  (((d * (q - p)) != ((u_int64_t)x - p)) || !!(d & ~0x3FUL)) {
                ifBLIN_QX1( "@%s:%s:%d: "
                            "Lost syn %"BLIN_X" i%"BLIN_X" c%"BLIN_X" j%"BLIN_X" s%"BLIN_X" f%02X"
                          , file
                          , func
                          , line
                          , BLIN_I(x)
                          , BLIN_I(&pgoblin_options->io[0])
                          , BLIN_I(&pgoblin_options->conn[0])
                          , BLIN_I(&pgoblin_options->job[0])
                          , BLIN_I(&pgoblin_options->style[0])
                          , -f
                          );
                d = 0x3F;
            }
            buf[0] = -f;
            buf[1] = d;
        }
out:
        if  (0 > (pgoblin_options->xtrace = open( pgoblin_options->ftrace
                                                , O_WRONLY | O_APPEND | O_EXLOCK
            )     )                             ) {
            ifBLIN_QW0("open(%s, WRONLY|APPEND|EXLOCK)", pgoblin_options->ftrace);
        } else {
            recobe_store(&buf[PGOBLIN_HEDLN], PGOBLIN_PIDLN, getpid());
            if  (!!(i = pgoblin_xwrt(pgoblin_options, buf, PGOBLIN_GOSHORT))) {
                ifBLIN_QW0("mife_writ %d <> %d", i, PGOBLIN_GOSHORT);
            }
            close(pgoblin_options->xtrace);
    }   }
#   undef blin_internal_flags
}

static char P[] = "Xftmqocjs";
u_int16_t   nums[65536];

static int
/*****************************************************************************
 **                                                                         **/
qdasprint(pgoblin_main *options, u_int32_t f, u_int32_t n) {               /**
 **                                                                         **
 *****************************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    u_int32_t l1 = options->flags & f;
    u_int32_t l2 = nums[(n >> 4) & 0x0FFFF] & (1 << (n & 0x0F));
    return((options->flags & PGOBLIN_T_ONLY) ? (l1 && l2) : (l1 || l1));
#   undef blin_internal_flags
}

static ssize_t
/*****************************************************************************
 **                                                                         **/
cload(pgoblin_main *options, mife_descriptor *md, off_t *offset) {         /**
 **                                                                         **
 *****************************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    ssize_t          rdr;
    ssize_t          trd;
    const u_char    *p;

    if  (!(p = winpoint(md, *offset, 1))) {
        ifBLIN_QW0("trace");
        trd = (errno == EFTYPE) ? -EX_DATAERR : -EX_IOERR;
        goto out;
    }
    if  (*p == 0x0FF) {
        ++*offset;
        trd = -1;
        goto out;
    }
    rdr = recobe_clen(*p);
    if  (!(p = winpoint(md, *offset, rdr))) {
        ifBLIN_QW0("trace");
        trd = (errno == EFTYPE) ? -EX_DATAERR : -EX_IOERR;
        goto out;
    }
    trd = recobe_cload(p, &rdr);
    if  (recobe_clen(*p) != rdr) {
        ifBLIN_QX0("recobe_cload <> recobe_clen");
    }
    *offset += rdr;
out:;
    return(trd);
#   undef blin_internal_flags
}

static int
/***********************************************************************************************
 **                                                                                           **/
txtprt(pgoblin_main *options, int dasprint, mife_descriptor *md, off_t *offset, char end) {  /**
 **                                                                                           **
 ***********************************************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    int              ex = EX_OK;
    ssize_t          trd;
    const u_char    *p;

    if  (-1 > (trd = cload(options, md, offset))) {
        ifBLIN_QW0("trace @ 0x%"BLIN_O"X", *offset);
        ex = -trd;
    } else if (0 > trd) {
        if  (dasprint && ('|' == end)) printf("|");
    } else {
        if  (!(p = winpoint(md, *offset, trd))) {
            ifBLIN_QW0("trace @ 0x%"BLIN_O"X", *offset);
            ex = (errno == EFTYPE) ? EX_DATAERR : EX_IOERR;
            goto out;
        }
        if  (dasprint) {
            printf("%c%.*s%c", (' ' == end) ? ' ' : '=', (int)trd, p, (' ' == end) ? 0 : end);
        }
        *offset += trd;
    }
out:;
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************
 *****************************************************************************
 **                                                                         **/
pgoblin_traceat( pgoblin_main *options                                     /**/
/**/           , mife_descriptor *(*popin)(pgoblin_main *options)          /**/
/**/           , const char *Tflag                                         /**/
/**/           ) {                                                         /**
 **                                                                         **
 *****************************************************************************
 *****************************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    u_char           t[pgoblin.binparam];
    u_char           s[pgoblin.symparam];
    int              ex = EX_OK;
    int              seqnum;
    off_t            offset;
    int              binreg;
    int              symreg;
    u_int16_t        cmd;
    mife_descriptor *md;
    const u_char    *p;
    u_int32_t        n;

    if  (!(md = popin(options))) {
        ifBLIN_QW0("popin mife_init");
        ex = EX_IOERR;
        goto out;
    }
    if  (!Tflag) {
        ifBLIN_QX0("no trace task");
        ex = EX_USAGE;
        errno = EINVAL;
        goto out;
    }
    bzero(nums, sizeof(nums));
    n = 0;
    for (char c = 0; ' ' < *(Tflag++);) {
        u_int32_t i;

        ifBLIN_QX4("%c @ %"BLIN_X, *Tflag, BLIN_I(Tflag));
        switch(*Tflag) {
        case 'O': options->flags |= PGOBLIN_T_ONLY;    break;
        case 'R': options->flags |= PGOBLIN_T_REGO;    break;
        case 'C': options->flags |= PGOBLIN_T_COMAND;  break;
        case 'E': options->flags |= PGOBLIN_T_EXECUTE; break;
        case 'M': options->flags |= PGOBLIN_T_MODULES; break;
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            i = n;
            n = strtoul(Tflag, (char **)&Tflag, 0);
            ifBLIN_QX4("N%u-%u %c", i, n, c);
            if  ('-' != c) i = n;
            c = *Tflag;
            ifBLIN_QX4("N%u-%u %c", i, n, c);
            for (; i <= n; ++i) nums[(i >> 4) & 0x0000FFFF] |= (1 << (i & 0x0000000F));
            ifBLIN_QX4("N%u %c", n, c);
            break;
    }   }
    offset = 0;
    for (seqnum = 0; ; ++seqnum) {
        fflush(stdout);
        if  (!(p = winpoint(md, offset, PGOBLIN_GOSHORT))) {
            if  (!mife_ctlsize(md)) break;
            ifBLIN_QW0("trace file");
            ex = (errno == EFTYPE) ? EX_DATAERR : EX_IOERR;
            goto out;
        }
        switch(*p) {
        case PGOBLIN_STYLE   : t[0] = P[8]; break;
        case PGOBLIN_JOB     : t[0] = P[7]; break;
        case PGOBLIN_CONNECT : t[0] = P[6]; break;
        case PGOBLIN_OUTSET  : t[0] = P[5]; break;
        case PGOBLIN_PQRESULT: t[0] = P[4]; break;
        case PGOBLIN_MIFEDES : t[0] = P[3]; break;
        case PGOBLIN_TEXTPARM: t[0] = P[2]; break;
        case PGOBLIN_PATHPARM: t[0] = P[1]; break;
        case 0               : t[0] = P[0]; break;
        default              : t[0] = 0;
        }
        if  (!!t[0]) {
            ++p;
            if  (qdasprint(options, PGOBLIN_T_REGO, seqnum)) {
                printf( "R%6d.%u P%c %c\n"
                      , seqnum
                      , (unsigned)recobe_load(p + 1, 3)
                      , t[0]
                      , pgoblin_regn[*p & 0x03F]
                      );
            }
            offset += PGOBLIN_GOSHORT;
            continue;
        }
        switch(*p & 0x0E0) {
        case 0xE0: {
                int dasprint = qdasprint(options, PGOBLIN_T_COMAND, seqnum);

        ;       if  (!(p = winpoint(md, offset, PGOBLIN_GOEXECU))) {
        ;           ifBLIN_QW0("trace file");
        ;           ex = (errno == EFTYPE) ? EX_DATAERR : EX_IOERR;
        ;           goto out;
        ;       }
        ;       cmd = (*(p) << 8) & 0x0100;
        ;       cmd |= *(p + 1);
        ;       bzero(t, pgoblin.binparam);
        ;       bzero(s, pgoblin.symparam);
        ;       for (symreg = 0; symreg < pgoblin.symparam; symreg++) {
        ;           if  (0 <= (binreg = pgoblin_breg(cmd, symreg))) {
        ;               s[symreg] = t[binreg] = p[PGOBLIN_GOSHORT + binreg] & 0x03F;
        ;       }   }
        ;       if  (dasprint) {
        ;           printf( "C%6d.%u (%ds %dd) N%d #%s %c%c%c%c%c%c\n"
                          , seqnum
                          , (unsigned)recobe_load(PGOBLIN_HEDLN + p, PGOBLIN_PIDLN)
                          , (int)recobe_load(PGOBLIN_GOCMD + p, sizeof(int32_t))
                          , (int)recobe_load(sizeof(int32_t) + PGOBLIN_GOCMD + p, sizeof(int32_t))
                          , (int)recobe_load(sizeof(int32_t) * 2 + PGOBLIN_GOCMD + p, sizeof(int32_t))
                          , pgoblin.a[cmd].str
                          , pgoblin_regn[s[0]]
                          , pgoblin_regn[s[1]]
                          , pgoblin_regn[s[2]]
                          , pgoblin_regn[s[3]]
                          , pgoblin_regn[s[4]]
                          , pgoblin_regn[s[5]]
                          );
        ;       }
        ;       offset += PGOBLIN_GOEXECU;
        ;       for (symreg = 0; symreg < pgoblin.symparam; symreg++) {
        ;           if  (0 <= (binreg = pgoblin_breg(cmd, symreg))) {
                        u_char ch = pgoblin.a[cmd].ch[binreg];

#ifndef NOTRACE_PATHPARM
        ;               if  (PGOBLIN_PATHPARM & ch) {
        ;                   ifBLIN_QX4("[%d]PATHPARM", binreg);
        ;                   if  (dasprint) {
        ;                       printf( "        %d %d [%c]%sPf"
                                      , binreg
                                      , symreg
                                      , pgoblin_regn[t[binreg] & 0x03F]
                                      , (ch & PGOBLIN_OUTSET) ? "Po|" : "   "
                                      );
        ;                   }
        ;                   if  (!!(ex = txtprt(options, dasprint, md, &offset, '~'))) {
        ;                       ifBLIN_QW0("trace Pf");
        ;                       goto out;
        ;                   }
        ;                   if  (dasprint) printf("\n");
        ;               }
#endif
#ifndef NOTRACE_TEXTPARM
        ;               if  (PGOBLIN_TEXTPARM & ch) {
        ;                   ifBLIN_QX4("[%d]TEXTPARM", binreg);
        ;                   if  (dasprint) {
        ;                       printf( "        %d %d [%c]%sPt"
                                      , binreg
                                      , symreg
                                      , pgoblin_regn[t[binreg] & 0x03F]
                                      , (ch & PGOBLIN_OUTSET) ? "Po|" : "   "
                                      );
        ;                   }
        ;                   if  (!!(ex = txtprt(options, dasprint, md, &offset, '~'))) {
        ;                       ifBLIN_QW0("trace Pt");
        ;                       goto out;
        ;                   }
        ;                   if  (dasprint) printf("\n");
        ;               }
#endif
#ifndef NOTRACE_MIFEDES
        ;               if  (PGOBLIN_MIFEDES  & ch) {
        ;                   ifBLIN_QX4("[%d]MIFEDES", binreg);
        ;               }
#endif
#ifndef NOTRACE_PQRESULT
        ;               if  (PGOBLIN_PQRESULT & ch) {
        ;                   ifBLIN_QX4("[%d]PQRESULT", binreg);
        ;                   if  (dasprint) {
        ;                       printf( "        %d %d [%c]%sPq"
                                      , binreg
                                      , symreg
                                      , pgoblin_regn[t[binreg] & 0x03F]
                                      , (ch & PGOBLIN_OUTSET) ? "Po|" : "   "
                                      );
        ;                   }
        ;                   if  (!(p = winpoint(md, offset, 1))) {
        ;                        ifBLIN_QW0("trace q");
        ;                        ex = (errno == EFTYPE) ? EX_DATAERR : EX_IOERR;
        ;                        goto out;
        ;                   }
        ;                   if  (*p == 0x0FF) {
        ;                       if  (dasprint) printf("\n");
        ;                       ++offset;
        ;                   } else {
                                int     row;
                                int     col;

        ;                       row = cload(options, md, &offset); /* 0 > ? XXXX*/
        ;                       col = cload(options, md, &offset); /* 0 > ? XXXX*/
        ;                       if  (dasprint) {
        ;                           printf(" %d rows x %d cols\n", row, col);
        ;                       }
        ;                       for (int r = 0; r < row; ++r) {
        ;                           for (int c = 0; c < col; ++c) {
        ;                               if  (!!(ex = txtprt(options, dasprint, md, &offset, '|'))) {
        ;                                   ifBLIN_QW0("trace Pq");
        ;                                   goto out;
        ;                           }   }
        ;                           if  (dasprint) printf("\n");
        ;               }   }   }
#endif
#ifndef NOTRACE_OUTSET
        ;               if  (PGOBLIN_OUTSET   & ch) {
        ;                   ifBLIN_QX4("[%d]OUTSET", binreg);
        ;               }
#endif
#ifndef NOTRACE_CONNECT
        ;               if  (PGOBLIN_CONNECT  & ch) {
        ;                   ifBLIN_QX4("[%d]CONNECT", binreg);
        ;               }
#endif
#ifndef NOTRACE_JOB
        ;               if  (PGOBLIN_JOB      & ch) {
        ;                   ifBLIN_QX4("[%d]JOB", binreg);
        ;               }
#endif
        ;               if  (PGOBLIN_STYLE    & ch) {
#ifndef NOTRACE_STYLE
        ;                   ifBLIN_QX4("[%d]STYLE", binreg);
#endif
        ;   }   }   }   }
        ;   break;
        case 0xC0: {
        ;       if  (!(p = winpoint(md, offset, PGOBLIN_GOEXECU))) {
        ;           ifBLIN_QW0("trace file");
        ;           ex = (errno == EFTYPE) ? EX_DATAERR : EX_IOERR;
        ;           goto out;
        ;       }
        ;       cmd = (*(p) << 8) & 0x0100;
        ;       cmd |= *(p + 1);
        ;       bzero(t, pgoblin.binparam);
        ;       bzero(s, pgoblin.symparam);
        ;       for (symreg = 0; symreg < pgoblin.symparam; symreg++) {
        ;           if  (0 <= (binreg = pgoblin_breg(cmd, symreg))) {
        ;               s[symreg] = t[binreg] = p[PGOBLIN_GOSHORT + binreg];
        ;       }   }
        ;       if  (qdasprint(options, PGOBLIN_T_EXECUTE, seqnum)) {
        ;           printf( "E%6d.%u (%ds %dd) N%d #%s %c%c%c%c%c%c\n"
                          , seqnum
                          , (unsigned)recobe_load(PGOBLIN_HEDLN + p, PGOBLIN_PIDLN)
                          , (int)recobe_load(p + PGOBLIN_GOCMD, sizeof(int32_t))
                          , (int)recobe_load(sizeof(int32_t) + p + PGOBLIN_GOCMD, sizeof(int32_t))
                          , (int)recobe_load(sizeof(int32_t) * 2 + p + PGOBLIN_GOCMD, sizeof(int32_t))
                          , pgoblin.a[cmd].str
                          , pgoblin_regn[s[0] & 0x03F]
                          , pgoblin_regn[s[1] & 0x03F]
                          , pgoblin_regn[s[2] & 0x03F]
                          , pgoblin_regn[s[3] & 0x03F]
                          , pgoblin_regn[s[4] & 0x03F]
                          , pgoblin_regn[s[5] & 0x03F]
                          );
        ;       }
        ;       offset += PGOBLIN_GOEXECU;
        ;   }
        ;   break;
        case 0xA0: {
                int dasprint = qdasprint(options, PGOBLIN_T_MODULES, seqnum);

        ;       if  (!(p = winpoint(md, offset, PGOBLIN_GOSHORT))) {
        ;           ifBLIN_QW0("trace file");
        ;           ex = (errno == EFTYPE) ? EX_DATAERR : EX_IOERR;
        ;           goto out;
        ;       }
        ;       if  (dasprint) {
        ;           printf( "M%6d.%u[%u]"
                          , seqnum
                          , (unsigned)recobe_load(p + PGOBLIN_HEDLN, PGOBLIN_PIDLN)
                          , *(p + 1)
                          );
        ;       }
        ;       offset += 1 + PGOBLIN_PIDLN;
        ;       for (int i = 0; i < 3; ++i) {
        ;           if  (!!(ex = txtprt(options, dasprint, md, &offset, ' '))) {
        ;               ifBLIN_QW0("trace mod");
        ;               goto out;
        ;       }   }
        ;       if  (dasprint) printf("\n");
        ;   }
        ;   break;
        default: {
        ;       ifBLIN_QX0("illegal trace file 0x%02X @ 0x%012"BLIN_O"X", *p, offset);
        ;       ex = EX_DATAERR;
        ;       errno = EILSEQ;
        ;       goto out;
    }   }   }
out:;
    return(ex);
#   undef blin_internal_flags
}
